/*
 * CONTRAT DE LICENCE DE LOGICIEL LIBRE CeCILL-B
 * https://cecill.info/licences/Licence_CeCILL-B_V1-fr.html
 */
package fr.unistra.sil.erp.back.db;

import fr.unistra.sil.erp.back.model.Category;
import fr.unistra.sil.erp.back.model.Item;
import fr.unistra.sil.erp.back.model.RegistryEntry;
import fr.unistra.sil.erp.back.model.Stock;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.time.LocalDateTime;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * SQLite implementation of the Database interface.
 * 
 * This allows the software to store and process information through
 * a SQLite database file.
 * 
 * @author BEAUVAIS ANTOINE
 */
public class DatabaseSQLiteImpl implements IDatabase {
    
    /**
     * SQLite's connection string for the JDBC driver.
     */
    private static final String CONNECTION_URL = "jdbc:sqlite:dev.db";
    
    /**
     * Query used to select all items stored in the database.
     */
    private static final String SQL_GETALLITEMS = 
            "SELECT ref, name, price, subscriberPrice, category FROM items";
    
    /**
     * Query used to get all items stored in the database from a specific
     * category.
     */
    private static final String SQL_GETITEMSFROMCATEGORY = SQL_GETALLITEMS +
            " WHERE category = ?";
    
    /**
     * Query used to retrieve all categories stored in the database.
     */
    private static final String SQL_GETCATEGORIES =
            "SELECT id, name FROM categories";
    
    /**
     * Query used to fetch all stocks stored in the database.
     */
    private static final String SQL_GETSTOCKS = 
            "SELECT s.id AS id, s.item AS item, " +
            "i.name AS name, s.quantity AS quantity FROM stocks s " +
            "INNER JOIN items i ON s.item = i.ref";
    
    private static final String SQL_GETSTOCKFORITEM = SQL_GETSTOCKS +
            " WHERE item = ?";
    
    /**
     * Query used to retrieve the registry.
     */
    private static final String SQL_GETREGISTRY =
            "SELECT r.id AS id, r.dt AS date, r.type AS type, " +
            "r.account_id AS account, r.debit AS debit, r.credit AS credit, " +
            "r.remarks AS remarks FROM Registry r";
    
    private static final String SQL_UPDATESTOCK =
            "UPDATE stocks SET quantity = ? WHERE item = ?";
    
    /**
     * SQL Connection object.
     */
    private Connection conn;
    
    /**
     * Class constructor.
     */
    public DatabaseSQLiteImpl()
    {
        try {
            conn = DriverManager.getConnection(CONNECTION_URL);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to connect to SQLite file.", ex);
        }
    }
    
    /**
     * Executes a standard query, without parameters.
     * @param query the string's query.
     * @return the result set for this query.
     */
    private ResultSet query(String query)
    {
        Statement stmt;
        ResultSet rs;
        try {
            stmt = this.conn.createStatement();
            rs = stmt.executeQuery(query);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to run query: " + query, ex);
            return null;
        }
        
        return rs;
    }

    /**
     * Retrieves all items from inside the database.
     * @return the list of items.
     */
    @Override
    public List<Item> getAllItems() {
        ResultSet rs = query(SQL_GETALLITEMS);
        if(rs == null)
            return null;
        
        List<Item> res = new ArrayList<>();
        try {
            while(rs.next())
            {
                Item item = new Item(rs.getInt("ref"), rs.getString("name"),
                                     rs.getBigDecimal("price"),
                                     rs.getBigDecimal("subscriberPrice"),
                                     rs.getInt("category"));
                res.add(item);
            }
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, null, ex);
            return null;
        }
        
        return res;
    }

    /**
     * Returns all items from the specified category.
     * @param category the category to filter the results with.
     * @return the list of items.
     */
    @Override
    public List<Item> getItemsFromCategory(int category) {
        
        PreparedStatement ps;
        try {
            ps = this.conn.prepareStatement(
                    SQL_GETITEMSFROMCATEGORY);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to connect to database.", ex);
            return null;
        }
        
        try {
            ps.setInt(1, category);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to set query parameter.", ex);
            return null;
        }
        
        ResultSet rs;
        try {
            rs = ps.executeQuery();
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to fetch items for category " +
                            category, ex);
            return null;
        }
        
        List<Item> res = new ArrayList<>();
        try {
            while(rs.next())
            {
                Item item = new Item(rs.getInt("ref"), rs.getString("name"),
                                     rs.getBigDecimal("price"),
                                     rs.getBigDecimal("subscriberPrice"),
                                     rs.getInt("category"));
                res.add(item);
            }
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to parse results.", ex);
            return null;
        }
        
        return res;
    }

    /**
     * Retrieves all categories from the database.
     * @return the list of categories.
     */
    @Override
    public List<Category> getCategories() {
        ResultSet rs = this.query(SQL_GETCATEGORIES);
        if(rs == null)
            return null;
        
        List<Category> res = new ArrayList<>();
        try {
            while(rs.next())
            {
                Category c = new Category(rs.getInt("id"), rs.getString("name"));
                res.add(c);
            }
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to fetch results.", ex);
            return null;
        }
        
        return res;
    }

    /**
     * Retrieves all stock entries from the database.
     * @return the list of stock entries.
     */
    @Override
    public List<Stock> getStocks() {
        ResultSet rs = this.query(SQL_GETSTOCKS);
        if(rs == null)
            return null;
        
        List<Stock> res = new ArrayList<>();
        try {
            while(rs.next())
            {
                Stock s = new Stock(rs.getInt("id"),
                        rs.getInt("item"), rs.getString("name"),
                        rs.getInt("quantity"));
                res.add(s);
            }
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to fetch results.", ex);
            return null;
        }
        
        return res;
    }

    /**
     * Returns the stock entry for the specified item.
     * @param itemId the item's ID.
     * @return the Stock entry.
     * @throws DatabaseResourceNotFoundException if the stock doesn't exist.
     */
    @Override
    public Stock getStockForItem(int itemId)
            throws DatabaseResourceNotFoundException {
        PreparedStatement ps;
        try {
            ps = this.conn.prepareStatement(SQL_GETSTOCKFORITEM);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to connect to database.", ex);
            return null;
        }
        
        try {
            ps.setInt(1, itemId);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to set query parameter.", ex);
            return null;
        }
        
        ResultSet rs;
        try {
            rs = ps.executeQuery();
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to fetch stock for item " +
                            itemId, ex);
            return null;
        }
        
        Stock s;
        try {
            if(rs.next())
                s = new Stock(rs.getInt("id"), rs.getInt("item"),
                                     rs.getString("name"),
                                     rs.getInt("quantity"));
            else
                throw new DatabaseResourceNotFoundException("Stock not found.");
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to parse results.", ex);
            return null;
        }
        
        return s;
    }

    /**
     * Updates the specified stock entry with the new quantity.
     * @param id the stock's ID.
     * @param quantity the associated quantity.
     * @throws DatabaseConnectionException when the connection to the DB fails.
     * @throws DatabaseUpdateException when the update fails.
     */
    @Override
    public void updateStock(int id, int quantity)
            throws DatabaseConnectionException, DatabaseUpdateException {
        
        PreparedStatement ps;
        try {
            ps = this.conn.prepareStatement(SQL_UPDATESTOCK);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to connect to database.", ex);
            throw new DatabaseConnectionException("Failed to connect to DB.");
        }
        
        try {
            ps.setInt(1, quantity);
            ps.setInt(2, id);
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to set query parameter.", ex);
            throw new DatabaseUpdateException("Failed to set query params.");
        }
        
        try {
            ps.execute();
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName())
                    .log(Level.SEVERE, "Failed to execute query.", ex);
            throw new DatabaseUpdateException("Failed to execute query.");
        }
    }

    @Override
    public List<RegistryEntry> getRegistry() throws DatabaseConnectionException {
        ResultSet rs = this.query(SQL_GETREGISTRY);
        if(rs == null)
            return null;
        
        List<RegistryEntry> res = new ArrayList<>();
        try {
            while(rs.next())
            {
                RegistryEntry r = new RegistryEntry(rs.getInt("id"),
                        rs.getInt("type"), new Date(rs.getInt("date") * 1000),
                        rs.getInt("account"),
                        rs.getBigDecimal("debit"), rs.getBigDecimal("credit"),
                        rs.getString("remarks"));
                res.add(r);
            }
        } catch (SQLException ex) {
            Logger.getLogger(DatabaseSQLiteImpl.class.getName()).log(
                    Level.SEVERE, "Failed to fetch results.", ex);
            return null;
        }
        
        return res;
    }
    
}
